import logging
import unittest.mock

import click.testing
import pytest
import requests

import mantik.cli.main as main
import mantik.cli.remote_file_service.unicore_file_service as cli_uc_fs
import mantik.remote_file_service.data_client as data_client
import mantik.testing as testing

GROUP_NAME = cli_uc_fs.GROUP_NAME


@pytest.fixture
def no_authentication():
    with unittest.mock.patch(
        "mantik.authentication.auth.get_valid_access_token",
        return_value="access-token",
    ) as mock:
        yield mock


@pytest.fixture
def mock_response():
    response = requests.Response()
    response.status_code = 200
    response._content = (
        b'{"dataRepositoryId": "test", '
        b'"dataRepositories": '
        b"[{"
        b'"dataRepositoryName": "test", '
        b'"dataRepositoryId": "3503c310-e2d1-4d3f-bd75-0a1679e48c78"}'
        b"]}"
    )
    return response


@pytest.fixture
def mock_mantik_api(no_authentication, mock_response):
    with unittest.mock.patch(
        "mantik.utils.mantik_api.client.send_request_to_mantik_api",
        return_value=mock_response,
    ) as mock:
        yield mock


@pytest.fixture
def fake_data_client(mock_mantik_api):
    with unittest.mock.patch(
        "mantik.remote_file_service.data_client.DataClient.from_env",
        return_value=data_client.DataClient(
            file_service=testing.remote_file_service.FakeUnicoreFileService(),
        ),
    ) as mock:
        yield mock


@pytest.fixture
def cli_test_runner():
    yield click.testing.CliRunner()


@pytest.fixture
def debug_caplog(caplog):
    caplog.set_level(logging.DEBUG)
    yield caplog


def test_copy_file(fake_data_client, debug_caplog, cli_test_runner):
    result = cli_test_runner.invoke(
        main.cli,
        [
            GROUP_NAME,
            "copy-file",
            "/fake-dir/fake.file",
            "remote:/fake-dir/fake.file",
        ],
    )

    assert result.exit_code == 0
    assert result.output == str(testing.remote_file_service.fake_file) + "\n"
    assert (
        "copy_file /fake-dir/fake.file remote:/fake-dir/fake.file"
        in debug_caplog.text
    )


def test_remove_file(fake_data_client, debug_caplog, cli_test_runner):
    result = cli_test_runner.invoke(
        main.cli,
        [
            GROUP_NAME,
            "remove-file",
            "remote:/fake-dir/fake.file",
        ],
    )

    assert result.exit_code == 0
    assert "remove_file remote:/fake-dir/fake.file" in debug_caplog.text


def test_create_file_if_not_exists(
    fake_data_client, debug_caplog, cli_test_runner
):
    result = cli_test_runner.invoke(
        main.cli,
        [
            GROUP_NAME,
            "create-file",
            "remote:/fake-dir/fake.file",
        ],
    )

    assert result.exit_code == 0
    assert result.output == str(testing.remote_file_service.fake_file) + "\n"
    assert (
        "create_file_if_not_exists remote:/fake-dir/fake.file"
        in debug_caplog.text
    )


def test_list_directory(fake_data_client, debug_caplog, cli_test_runner):
    result = cli_test_runner.invoke(
        main.cli,
        [
            GROUP_NAME,
            "list-directory",
            "remote:/fake-dir/fake.file",
        ],
    )

    assert result.exit_code == 0
    assert (
        result.output
        == str(testing.remote_file_service.fake_directory.children) + "\n"
    )
    assert "list_directory remote:/fake-dir/fake.file" in debug_caplog.text


def test_create_directory(fake_data_client, debug_caplog, cli_test_runner):
    result = cli_test_runner.invoke(
        main.cli,
        [
            GROUP_NAME,
            "create-directory",
            "remote:/fake-dir",
        ],
    )

    assert result.exit_code == 0
    assert (
        result.output == str(testing.remote_file_service.fake_directory) + "\n"
    )
    assert "create_directory remote:/fake-dir" in debug_caplog.text


def test_remove_directory(fake_data_client, debug_caplog, cli_test_runner):
    result = cli_test_runner.invoke(
        main.cli,
        [
            GROUP_NAME,
            "remove-directory",
            "remote:/fake-dir",
        ],
    )

    assert result.exit_code == 0
    assert "remove_directory remote:/fake-dir" in debug_caplog.text


def test_directory_file(fake_data_client, debug_caplog, cli_test_runner):
    result = cli_test_runner.invoke(
        main.cli,
        [
            GROUP_NAME,
            "copy-directory",
            "/fake-dir",
            "remote:/fake-dir",
        ],
    )

    assert result.exit_code == 0
    assert (
        result.output == str(testing.remote_file_service.fake_directory) + "\n"
    )
    assert "copy_directory /fake-dir remote:/fake-dir" in debug_caplog.text


def test_exists(fake_data_client, debug_caplog, cli_test_runner):
    result = cli_test_runner.invoke(
        main.cli,
        [
            GROUP_NAME,
            "exists",
            "remote:/fake-dir/fake.file",
        ],
    )

    assert result.exit_code == 0
    assert result.output == "True\n"


def test_user(fake_data_client, debug_caplog, cli_test_runner):
    result = cli_test_runner.invoke(
        main.cli,
        [
            GROUP_NAME,
            "user",
        ],
    )

    assert result.exit_code == 0
    assert result.output == "FAKE-USER\n"


def test_cli_operation_with_wrong_connection_id(
    fake_data_client, debug_caplog, cli_test_runner
):
    result = cli_test_runner.invoke(
        main.cli,
        [
            GROUP_NAME,
            "copy-file",
            "/fake-dir/fake.file",
            "remote:/fake-dir/fake.file",
            "--connection-id",
            "wrong-id-type",
        ],
    )
    assert "Error: Invalid value for '--connection-id'" in result.output
    assert result.exit_code == 2



import pathlib
import typing as t
import uuid

import mantik.authentication.auth
import mantik.runs.schemas as schemas
import mantik.utils.mantik_api as mantik_api


def submit_run(
    project_id: uuid.UUID,
    name: str,
    experiment_repository_id: uuid.UUID,
    code_repository_id: uuid.UUID,
    branch: t.Optional[str],
    commit: t.Optional[str],
    data_repository_id: t.Optional[uuid.UUID],
    data_branch: t.Optional[str],
    data_commit: t.Optional[str],
    mlflow_mlproject_file_path: str,
    entry_point: str,
    mlflow_parameters: dict,
    backend_config: t.Union[pathlib.Path, str, dict],
    connection_id: t.Optional[uuid.UUID],
    compute_budget_account: t.Optional[str],
):
    """

    Parameters
    ----------
    project_id : ID of the project to which the run should be linked
    name : Name of the Run
    experiment_repository_id : ID of the experiment repository
        to which the run should be linked
    code_repository_id : ID of the code repository
        where the mlproject is located
    branch : Name of the code repository's branch
    commit : Name of the code repository's commit (has precedence over branch)
    data_repository_id : ID of the data repository
        where the data is located, this is optional
    data_branch : Data branch to checkout. Defaults to newest.
    data_commit : Data commit to checkout. Takes precedence over data_branch.
    mlflow_mlproject_file_path : Path in your code directory
        to the MLproject file
    entry_point : entry point name
    mlflow_parameters : Mlflow parameters present in your entry point
    backend_config : path to backend configuration file
        relative to path from where this command is called.
        It has to be a local file.
    connection_id : Mantik Connection UUID
    compute_budget_account : Name of the compute budget account on HPC

    Returns
    -------
    Response from the mantik API that contains the run id
    """
    run_configuration = schemas.RemoteRunConfiguration(
        name=name,
        experiment_repository_id=experiment_repository_id,
        data_repository_id=data_repository_id,
        data_branch=data_branch,
        data_commit=data_commit,
        code_repository_id=code_repository_id,
        branch=branch,
        commit=commit,
        connection_id=connection_id,
        compute_budget_account=compute_budget_account,
        mlflow_mlproject_file_path=mlflow_mlproject_file_path,
        entry_point=entry_point,
        mlflow_parameters=mlflow_parameters,
        backend_config=backend_config,
    )

    token = mantik.authentication.auth.get_valid_access_token()
    response = mantik_api.run.submit_run(
        project_id=project_id,
        submit_run_data=run_configuration.to_post_payload(),
        token=token,
    )
    return response



import pytest
import requests

import mantik.utils.mantik_api.code_repository as code_api


@pytest.mark.parametrize(
    ("status_code", "expected"), [(200, None), (404, requests.HTTPError())]
)
def test_get_model(
    mock_mantik_api_request,
    info_caplog,
    status_code,
    expected,
    sample_code_repository_id,
    sample_code_repository,
    project_id,
):
    sample_schema = sample_code_repository.to_json()

    with mock_mantik_api_request(
        method="GET",
        end_point=f"/projects/{str(project_id)}/"
        f"code/{str(sample_code_repository_id)}",
        status_code=status_code,
        json_response=sample_schema,
        expected_error=expected,
    ) as (m, error):
        retrieved_model = code_api.get_one(
            project_id=project_id,
            code_repository_id=sample_code_repository_id,
            token="test_token",
        )
        assert retrieved_model == code_api.CodeRepository.from_json(
            sample_schema
        )
    if error:
        assert any(
            "Call to Mantik API" in message for message in info_caplog.messages
        )



import dataclasses
import io
import logging
import os
import pathlib
import sys
import tempfile
import typing as t
import uuid

import mantik.authentication.auth
import mantik.authentication.tokens as tokens
import mantik.runs.schemas as schemas
import mantik.utils.env_vars as env_vars
import mantik.utils.mantik_api as mantik_api
import mantik.utils.mantik_api.code_repository as code_api
import mantik.utils.mantik_api.connection as connection_api
import mantik.utils.mantik_api.experiment_repository as experiment_api
from mantik.utils.other import construct_git_clone_uri

logger = logging.getLogger(__name__)


def start_local_run(
    project_id: uuid.UUID,
    name: str,
    experiment_repository_id: uuid.UUID,
    code_repository_id: uuid.UUID,
    branch: t.Optional[str],
    commit: t.Optional[str],
    data_repository_id: t.Optional[uuid.UUID],
    mlflow_mlproject_file_path: str,
    entry_point: str,
    data_branch: t.Optional[str],
    data_commit: t.Optional[str],
    data_target_dir: t.Optional[str],
    env_manager: t.Optional[str],
    mlflow_parameters: dict,
):
    """

    Parameters
    ----------
    project_id : ID of the project to which the run should be linked
    name : Name of the Run
    experiment_repository_id : ID of the experiment repository
        to which the run should be linked
    code_repository_id : ID of the code repository
        where the mlproject is located
    branch : Name of the code repository's branch
    commit : Name of the code repository's commit (has precedence over branch)
    data_repository_id : ID of the data repository
        where the data is located, this is optional
    mlflow_mlproject_file_path : Path in your code directory
        to the MLproject file
    entry_point : entry point name
    mlflow_parameters : Mlflow parameters present in your entry point
    data_branch : Data branch to checkout. Defaults to newest.
    data_commit : Data commit to checkout. Takes precedence over data_branch.
    env_manager: Which environment manager to use to handle installing run
    dependencies. Could be Local, Conda or Virtualenv
    data_target_dir : Relative path to directory
        where the data will be stored (from code root)

    Returns
    -------
    Response from the mantik API that contains the run id
    """
    token = mantik.authentication.auth.get_valid_access_token()
    run(
        data=schemas.RunConfiguration(
            name=name,
            experiment_repository_id=experiment_repository_id,
            code_repository_id=code_repository_id,
            branch=branch,
            commit=commit,
            data_repository_id=data_repository_id,
            data_branch=data_branch,
            data_commit=data_commit,
            mlflow_mlproject_file_path=mlflow_mlproject_file_path,
            entry_point=entry_point,
            mlflow_parameters=mlflow_parameters,
            backend_config={},
        ),
        data_target_dir=data_target_dir,
        project_id=project_id,
        mantik_token=token,
        env_manager=env_manager,
    )


@dataclasses.dataclass
class LocalRunOutput:
    run_id: uuid.UUID
    exception: t.Optional[BaseException] = None


class LocalRunManager:
    """
    The purpose of this class is so that the two functions inside it
    can be more easily monkey-patched together,
    feel free to improve this by monkeypatching the functions
    without the need of an extra class to clean up the code
    """

    def __init__(self):
        pass

    @classmethod
    def clone_git_repo(
        cls,
        uri: str,  # public or private
        branch: str,
        target_directory: str,
    ) -> None:
        try:
            import git
        except ModuleNotFoundError as e:
            raise RuntimeError(
                "Please install git module: required to execute local runs"
            ) from e

        logger.debug("Cloning repo...")

        git.Repo.clone_from(uri, target_directory, branch=branch)

        logger.debug("Repo cloned.")

    @staticmethod
    def start_local_run(
        mlflow_experiment_id: str,
        data: schemas.RunConfiguration,
        project_id: uuid.UUID,
        mantik_token: str,
        uri: str,
        env_manager: str = "local",
    ) -> LocalRunOutput:
        # Lazy import so that mantik can be used without mlflow
        import mlflow
        import mantik.mlflow
        import mantik.runs.mlflow_monkeypatch

        # NOTE: If additional mlflow functions need to be monkeypatched,
        # add them to the mlflow_monkeypatch module.
        #
        # - Monkeypatched run_mlflow_run_cmd and run_entry_point
        #   to capture output.
        #
        # Reason: We want to display the logs of a run in the Mantik UI.
        # Since mlflow does not support saving the logs of subprocesses
        # (this might change: https://github.com/mlflow/mlflow/issues/863,
        # but it may only apply to the main process logs),
        # we need to modify how stdout and stderr of the subprocess are handled
        # to capture them for display in the Mantik UI.

        mlflow.projects.backend.local._run_mlflow_run_cmd = (
            mantik.runs.mlflow_monkeypatch.run_mlflow_run_cmd
        )
        mlflow.projects.backend.local._run_entry_point = (
            mantik.runs.mlflow_monkeypatch.run_entry_point
        )

        logger.info(f"Starting mlflow run: '{data.name}' ...")
        mlflow.set_tracking_uri(
            os.getenv(
                mantik.utils.mlflow.TRACKING_URI_ENV_VAR,
                mantik.utils.mlflow.DEFAULT_TRACKING_URI,
            )
        )
        with mantik.mlflow.start_run(
            experiment_id=mlflow_experiment_id,
            run_name=data.name,
        ) as active_run:
            data.mlflow_run_id = active_run.info.run_id
            run_id = save_run_data(
                data=data, project_id=project_id, mantik_token=mantik_token
            )
            mantik_api.run.update_run_status(
                project_id=project_id,
                token=mantik_token,
                status=active_run.info.status,
                run_id=run_id,
            )
        try:
            _run = mlflow.run(
                uri=uri,
                backend="local",
                experiment_id=mlflow_experiment_id,
                run_name=data.name,
                entry_point=data.entry_point,
                parameters=data.mlflow_parameters,
                backend_config={
                    mantik.mlflow.get_local_backend_config_run_id_env_var(): active_run.info.run_id  # noqa: E501
                },
                env_manager=env_manager,
                build_image=True,
                docker_args={
                    "e": f"{mantik.utils.mlflow.TRACKING_TOKEN_ENV_VAR}={os.environ[mantik.utils.mlflow.TRACKING_TOKEN_ENV_VAR]}"  # noqa: E501
                },
            )
        except KeyboardInterrupt as e:
            logger.warning(
                "Keyboard interrupt, setting Mantik run status to KILLED"
            )
            mantik_api.run.update_run_status(
                project_id=project_id,
                token=mantik_token,
                status="KILLED",
                run_id=run_id,
            )
            return LocalRunOutput(run_id, exception=e)
        # MlflowException is raised when something goes wrong with mlflow
        # FileNotFoundError is raised when pyenv or conda are not found
        # This catches any other exception
        # so that we can in any case push the logs and update the status
        except Exception as e:
            logger.warning(
                "Failed to execute run, setting Mantik run status to FAILED"
            )
            mantik_api.run.update_run_status(
                project_id=project_id,
                token=mantik_token,
                status="FAILED",
                run_id=run_id,
            )
            return LocalRunOutput(run_id, exception=e)
        logger.debug("Run finished successfully")
        mantik_api.run.update_run_status(
            project_id=project_id,
            token=mantik_token,
            status=_run.get_status(),
            run_id=run_id,
        )
        return LocalRunOutput(run_id)


def fetch_code_and_experiment(
    project_id: uuid.UUID,
    code_repository_id: uuid.UUID,
    experiment_repository_id: uuid.UUID,
    mantik_token: str,
) -> t.Tuple[code_api.CodeRepository, experiment_api.ExperimentRepository]:
    logger.debug("Fetching code and experiment from Mantik API...")
    code = code_api.get_one(
        code_repository_id=code_repository_id,
        project_id=project_id,
        token=mantik_token,
    )
    experiment = experiment_api.get_one(
        experiment_repository_id=experiment_repository_id,
        project_id=project_id,
        token=mantik_token,
    )
    logger.debug(f"Fetched {code} and {experiment}")
    logger.debug(f"Code repo uri: {code.uri}")
    return code, experiment


def save_run_data(
    data: schemas.RunConfiguration, project_id: uuid.UUID, mantik_token: str
) -> uuid.UUID:
    logger.debug(f"Saving data... {data}")

    response = mantik_api.run.save_run(
        project_id=project_id,
        run_data=data.to_post_payload(),
        token=mantik_token,
    )

    logger.debug("Results saved to Mantik API")
    return uuid.UUID(response.json()["runId"])


def run(
    data: schemas.RunConfiguration,
    project_id: uuid.UUID,
    mantik_token: str,
    data_target_dir: t.Optional[str] = None,
    run_manager: LocalRunManager = LocalRunManager(),
    env_manager: t.Optional[str] = "local",
):
    code, experiment = fetch_code_and_experiment(
        project_id=project_id,
        code_repository_id=data.code_repository_id,
        experiment_repository_id=data.experiment_repository_id,
        mantik_token=mantik_token,
    )

    data.name = experiment_api.get_unique_run_name(
        experiment_repository_id=data.experiment_repository_id,
        project_id=project_id,
        token=mantik_token,
        run_name=data.name,
    )
    git_connection = (
        connection_api.get(
            user_id=uuid.UUID(tokens.get_user_id_from_token(mantik_token)),
            connection_id=uuid.UUID(code.connection_id),
            token=mantik_token,
        )
        if code.connection_id
        else None
    )
    uri = code.uri
    if (
        git_connection
    ):  # means that the git is private, therefore token needs to be injected
        uri = construct_git_clone_uri(
            uri=uri,
            git_access_token=git_connection.token,
            platform=code.platform,
        )
    with tempfile.TemporaryDirectory() as temp_dir:
        run_manager.clone_git_repo(
            uri=uri,
            branch=data.commit or data.branch,
            target_directory=temp_dir,
        )

        with mantik.utils.env.env_vars_overwrite_temporarily(
            get_env_vars_for_run(
                run_config=data,
                mantik_token=mantik_token,
                project_id=str(project_id),
                data_target_dir=data_target_dir,
            )
        ):
            start_local_run_output, captured_output = capture_and_print_output(
                func=run_manager.start_local_run,
                uri=path_directory_of_mlproject_file(
                    str(
                        pathlib.Path(temp_dir).joinpath(
                            data.mlflow_mlproject_file_path
                        )
                    )
                ),
                data=data,
                mlflow_experiment_id=experiment.mlflow_experiment_id,
                mantik_token=mantik_token,
                project_id=project_id,
                env_manager=env_manager,
            )

        mantik_api.run.update_logs(
            project_id=project_id,
            token=mantik_token,
            logs=captured_output,
            run_id=start_local_run_output.run_id,
        )

        if start_local_run_output.exception:
            raise start_local_run_output.exception


def path_directory_of_mlproject_file(mlproject_file_path: str) -> str:
    return str(pathlib.Path(mlproject_file_path).parent)


class Tee(io.StringIO):
    """
    Like the tee linux command that prints to stdout
    and save to a/multiple file the output of a command.
    This class takes n streams and populates them with the same data.
    A stream could be sys.stdout or sys.stderr, to keep printing to them.
    """

    def __init__(self, *streams):
        super().__init__()
        self.streams = streams

    def write(self, data):
        for stream in self.streams:
            stream.write(data)
        super().write(data)

    def flush(self):
        for stream in self.streams:
            stream.flush()
        super().flush()


def capture_and_print_output(func, *args, **kwargs) -> t.Tuple[t.Any, str]:
    original_stdout = sys.stdout
    original_stderr = sys.stderr
    output = io.StringIO()

    tee_stdout = Tee(sys.stdout, output)
    tee_stderr = Tee(sys.stderr, output)
    sys.stdout = tee_stdout
    sys.stderr = tee_stderr

    try:
        result = func(*args, **kwargs)
    finally:
        # Restore original stdout and stderr
        sys.stdout = original_stdout
        sys.stderr = original_stderr

    captured_output = output.getvalue()
    return result, captured_output


def get_env_vars_for_run(
    run_config: schemas.RunConfiguration,
    mantik_token: str,
    project_id: str,
    data_target_dir: t.Optional[str] = None,
) -> dict:
    entrypoint_env_vars = {
        env_vars.MANTIK_ACCESS_TOKEN_ENV_VAR: mantik_token,
        env_vars.PROJECT_ID_ENV_VAR: project_id,
    }

    if not run_config.data_repository_id:
        return entrypoint_env_vars

    entrypoint_env_vars[env_vars.DATA_REPOSITORY_ID_ENV_VAR] = str(
        run_config.data_repository_id
    )

    if run_config.data_branch:
        entrypoint_env_vars[
            env_vars.DATA_REPOSITORY_BRANCH_ENV_VAR
        ] = run_config.data_branch

    if run_config.data_commit:
        entrypoint_env_vars[
            env_vars.DATA_REPOSITORY_COMMIT_ENV_VAR
        ] = run_config.data_commit

    if not data_target_dir:
        raise ValueError(
            "Required to specify `data_target_dir` if data used for local run!"
        )

    entrypoint_env_vars[env_vars.TARGET_DIR_ENV_VAR] = data_target_dir

    return entrypoint_env_vars



import argparse
import ast
import typing as t

import mantik.authentication.auth


def dict_from_list(parameters: t.List[str]) -> t.Dict:
    return dict([_parse_parameter_from_string(p) for p in parameters])


def _parse_parameter_from_string(parameter: str) -> t.Tuple[str, t.Any]:
    key, value = parameter.split("=", 1)
    return key, _parse_value(value)


def _parse_value(value: t.Any) -> t.Any:
    try:
        return ast.literal_eval(value)
    except (ValueError, SyntaxError):
        # If value is a string, `astr.literal_eval` raises ValueError
        # and in some cases a SyntaxError.
        try:
            return ast.literal_eval(f"'{value}'")
        except (ValueError, SyntaxError):
            raise argparse.ArgumentTypeError(f"Unable to parse {value}")


def access_token_from_env_vars() -> str:
    return mantik.authentication.auth.get_valid_access_token()



import functools
import json
import pathlib
import uuid

import mlflow.projects as projects
import pytest
import pyunicore.client

import mantik.config.core as core
import mantik.testing as testing
import mantik.utils as mantik_utils
import mantik.utils.credentials as _credentials
import mantik_compute_backend.backend as backend

FILE_PATH = pathlib.Path(__file__).parent

ALL_ENV_VARS = [
    _credentials.UNICORE_USERNAME_ENV_VAR,
    _credentials.UNICORE_PASSWORD_ENV_VAR,
    core.COMPUTE_BUDGET_ACCOUNT_ENV_VAR,
]


class TestUnicoreBackend:
    def test_run(
        self, monkeypatch, example_project_path, tmp_path, env_vars_set, caplog
    ):
        monkeypatch.setattr(
            pyunicore.client,
            "Transport",
            testing.pyunicore.FakeTransport,
        )
        fake_client_with_successful_login = functools.partial(
            testing.pyunicore.FakeClient,
            login_successful=True,
        )
        monkeypatch.setattr(
            pyunicore.client,
            "Client",
            fake_client_with_successful_login,
        )
        backend_config_path = (
            example_project_path / "compute-backend-config.json"
        )
        with open(backend_config_path) as infile:
            backend_config = json.load(infile)
        # Following env vars are set by MLflow before running a project.
        backend_config[
            projects.PROJECT_STORAGE_DIR
        ] = example_project_path.as_posix()

        # Point MLFLOW_TRACKING_URI to a temporary directory
        tracking_uri = tmp_path / "mlruns"

        with env_vars_set(
            {
                # Following env vars must be set for the config.
                **{key: "test-val" for key in ALL_ENV_VARS},
                mantik_utils.mlflow.TRACKING_URI_ENV_VAR: tracking_uri.as_uri(),
                mantik_utils.mlflow.ACTIVE_RUN_ID_ENV_VAR: str(
                    uuid.uuid4().hex
                ),
            }
        ):
            submitted_run = backend.Backend().run(
                project_uri=example_project_path.as_posix(),
                entry_point="main",
                params={"print": "test"},
                version=None,
                backend_config=backend_config,
                tracking_uri=None,
                experiment_id=None,
            )

            assert submitted_run._job._job.started


def test_submit_job_and_upload_input_files(
    monkeypatch, example_project_path, example_project, example_unicore_config
):
    monkeypatch.setattr(
        pyunicore.client,
        "Transport",
        testing.pyunicore.FakeTransport,
    )
    fake_client_with_successful_login = functools.partial(
        testing.pyunicore.FakeClient,
        login_successful=True,
    )
    monkeypatch.setattr(
        pyunicore.client,
        "Client",
        fake_client_with_successful_login,
    )
    parameters = {"print": "test"}
    apptainer_image = example_project_path / "mantik-test.sif"
    backend._submit_job_and_upload_input_files(
        project=example_project,
        entry_point="main",
        parameters=parameters,
        project_dir=example_project_path,
        storage_dir="",
        input_files=[apptainer_image],
        config=example_unicore_config,
        run_id="example_run_id",
    )

    assert pathlib.Path(example_project_path / "mantik.sh").exists()
    pathlib.Path(example_project_path / "mantik.sh").unlink()


def test_create_job_description_apptainer(
    example_project_path, example_project, example_unicore_config
):
    expected = {
        "Executable": 'source $(dirname "$(realpath "$0")")/mantik.sh',
        "Arguments": [],
        "Project": "test-project",
        "Resources": {"Queue": "batch"},
        "RunUserPrecommandOnLoginNode": True,
        "RunUserPostcommandOnLoginNode": True,
        "Stderr": "mantik.log",
        "Stdout": "mantik.log",
    }

    result = example_unicore_config.to_unicore_job_description(
        bash_script_name="mantik.sh"
    )

    # Environment contains additional MLFLOW env vars,
    # which depend on the execution environment
    expected_environment = {
        "SRUN_CPUS_PER_TASK": 100,
        "MANTIK_WORKING_DIRECTORY": "$UC_WORKING_DIRECTORY",
    }
    actual_environment = result.pop("Environment")
    assert all(
        actual_environment[key] == value
        for key, value in expected_environment.items()
    )

    assert result == expected


def test_create_job_description_python(
    example_project_path, example_project, example_unicore_config_for_python
):
    expected = {
        "Executable": 'source $(dirname "$(realpath "$0")")/mantik.sh',
        "Arguments": [],
        "Project": "test-project",
        "Resources": {"Queue": "batch"},
        "RunUserPrecommandOnLoginNode": True,
        "RunUserPostcommandOnLoginNode": True,
        "Stderr": "mantik.log",
        "Stdout": "mantik.log",
    }
    result = example_unicore_config_for_python.to_unicore_job_description(
        bash_script_name="mantik.sh"
    )

    # Environment contains additional MLFLOW env vars,
    # which depend on the execution environment
    expected_environment = {"MANTIK_WORKING_DIRECTORY": "$UC_WORKING_DIRECTORY"}
    actual_environment = result.pop("Environment")
    assert all(
        actual_environment[key] == value
        for key, value in expected_environment.items()
    )

    assert result == expected


@pytest.mark.parametrize(
    ("entry_point", "config", "expected"),
    [
        (
            "main",
            "example_unicore_config_for_python",
            (
                "echo 'Submitted bash script'\n"
                'cat "$(realpath "$0")"\n'
                "precommand compute node && source /venv/bin/activate "
                "&& python main.py "
                "whatever && postcommand compute node"
            ),
        ),
        (
            "main",
            "example_unicore_config",
            (
                "echo 'Submitted bash script'\n"
                'cat "$(realpath "$0")"\n'
                "srun apptainer run mantik-test.sif python main.py whatever"
            ),
        ),
        (
            "multi-line",
            "example_unicore_config",
            (
                "echo 'Submitted bash script'\n"
                'cat "$(realpath "$0")"\n'
                "srun apptainer run mantik-test.sif python main.py whatever  "
                "--o option1  --i option2"
            ),
        ),
    ],
)
def test_to_bash_script(
    request,
    example_project_path,
    example_project,
    entry_point,
    config,
    expected,
):
    config = request.getfixturevalue(config)

    entry_point = example_project.get_entry_point(entry_point)
    parameters = {"print": "whatever"}
    storage_dir = "test-dir"

    arguments = backend._create_arguments(
        entry_point=entry_point, parameters=parameters, storage_dir=storage_dir
    )

    result = config.to_bash_script(arguments)

    assert result == expected



import logging
import os
import pathlib
import subprocess
import typing as t
import uuid

import git

import mantik.authentication.tokens as tokens
import mantik.utils.env_vars as env_vars
import mantik.utils.mantik_api.connection
import mantik.utils.mantik_api.data_repository
import mantik.utils.other as utils_other

logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO)


def data_download(
    project_id: t.Optional[uuid.UUID] = None,
    data_repository_id: t.Optional[uuid.UUID] = None,
    branch: t.Optional[str] = None,
    commit: t.Optional[str] = None,
    target_dir: t.Optional[pathlib.Path] = None,
    mantik_access_token: t.Optional[str] = None,
) -> str:
    """Downloads a GIT based data repository, along with its DVC files.

    Is meant to be used in the training script.
    `import mantik`
    `mantik.data_download()`
    """
    try:
        project_id = project_id or os.environ[env_vars.PROJECT_ID_ENV_VAR]
        data_repository_id = (
            data_repository_id
            or os.environ[env_vars.DATA_REPOSITORY_ID_ENV_VAR]
        )
        target_dir = target_dir or pathlib.Path(
            os.environ[env_vars.TARGET_DIR_ENV_VAR]
        )
        mantik_access_token = (
            mantik_access_token
            or os.environ[env_vars.MANTIK_ACCESS_TOKEN_ENV_VAR]
        )
    except KeyError as e:
        raise RuntimeError(
            f"mantik.data_download() requires environment variable {e}"
        )

    branch = branch or os.getenv(env_vars.DATA_REPOSITORY_BRANCH_ENV_VAR)
    commit = commit or os.getenv(env_vars.DATA_REPOSITORY_COMMIT_ENV_VAR)
    return download_data_repository(
        project_id=project_id,
        data_repository_id=data_repository_id,
        checkout=commit or branch,
        target_dir=target_dir,
        token=mantik_access_token,
    )


def download_data_repository(
    project_id: uuid.UUID,
    data_repository_id: uuid.UUID,
    checkout: t.Optional[str],
    target_dir: pathlib.Path,
    token: str,
) -> str:
    """Downloads a GIT based data repository, along with its DVC files."""
    data_repository_details = mantik.utils.mantik_api.data_repository.get_one(
        project_id=project_id,
        data_repository_id=data_repository_id,
        token=token,
    )

    git_uri = data_repository_details.uri

    if data_repository_details.connection_id is not None:
        git_connection = mantik.utils.mantik_api.connection.get(
            user_id=uuid.UUID(tokens.get_user_id_from_token(token)),
            connection_id=data_repository_details.connection_id,
            token=token,
        )
        git_uri = utils_other.construct_git_clone_uri(
            uri=git_uri,
            git_access_token=git_connection.token,
            platform=data_repository_details.platform,
        )

    git_clone_with_checkout(
        git_uri=git_uri,
        checkout=checkout,
        target_dir=target_dir,
    )

    if (
        not data_repository_details.is_dvc_enabled
        or not data_repository_details.dvc_connection_id
    ):
        return f"Cloned to {target_dir}"

    connection = mantik.utils.mantik_api.connection.get(
        user_id=uuid.UUID(tokens.get_user_id_from_token(token)),
        connection_id=data_repository_details.dvc_connection_id,
        token=token,
    )

    if connection.connection_provider == "S3":
        dvc_pull_with_aws_credentials(
            aws_access_key_id=connection.login_name,
            aws_secret_access_key=connection.password,
            target_dir=target_dir,
        )
    else:
        raise ValueError(
            "Connection provider not supported by our DVC backend."
        )

    return f"Cloned to {target_dir} with DVC"


def git_clone_with_checkout(
    git_uri: str, checkout: t.Optional[str], target_dir: pathlib.Path
):
    """Make target folder, git clone, and checkout a specific commit."""

    # clone the git repository
    logger.info("Cloning the git data repository...")
    target_dir.mkdir(parents=True, exist_ok=True)
    repo = git.Repo.clone_from(git_uri, target_dir)

    # checkout desired version
    if not checkout:
        return

    with utils_other.temp_chdir(target_dir):
        repo.git.checkout(checkout)


def dvc_pull_with_aws_credentials(
    aws_access_key_id: str,
    aws_secret_access_key: str,
    target_dir: t.Optional[pathlib.Path] = None,
):
    """Perform DVC pull using S3 as a DVC backend"""
    verify_dvc_is_installed()
    logger.info("Pulling files from the DVC backend...")

    with utils_other.temp_chdir(target_dir or os.getcwd()):
        subprocess.run(
            ["dvc", "pull"],
            env={
                "PATH": os.environ["PATH"],
                "AWS_ACCESS_KEY_ID": aws_access_key_id,
                "AWS_SECRET_ACCESS_KEY": aws_secret_access_key,
            },
        )


def verify_dvc_is_installed():
    try:
        logger.info("Verifying dvc installation...")
        subprocess.run(["dvc", "--version"])
    except FileNotFoundError:
        raise RuntimeError(
            "DVC is not installed. Please refer to https://dvc.org/doc/install"
        )
    else:
        logger.info("DVC is installed.")



import dataclasses
import itertools
import logging
import os
import pathlib
import time
import typing as t

import mlflow.entities
import pandas as pd

import mantik
import mantik.testing.stress as stress

logger = logging.getLogger()
stress.setup_logger(logger)


@dataclasses.dataclass
class Results(stress.ResultsBase):
    total_size: int
    n_files: int
    duration_per_file: float = None
    duration_per_mb: float = None

    def __post_init__(self):
        self.duration_per_file = self.duration / self.n_files
        self.duration_per_mb = self.duration / self.total_size
        logger.info(
            (
                "Sending %s files with total size %s MB via "
                "mlflow.%s %s %s seconds, "
                "average duration per MB was %s seconds, "
                "average duration per file was %s seconds, "
            ),
            self.n_files,
            self.total_size,
            self.mlflow_method,
            "failed after" if self.failed else "succeeded in",
            self.duration,
            self.duration_per_mb,
            self.duration_per_file,
        )

    def to_table_row(self) -> str:
        return (
            f"| {self.total_size} "
            f"| {self.n_files} "
            f"| {self.mlflow_method} "
            f"| {self.duration} "
            f"| {self.duration_per_mb} "
            f"| {self.duration_per_file} "
            f"| {not self.failed} "
            f"|\n"
        )

    def insert_to_df(self, df: pd.DataFrame) -> None:
        row = df.loc[self.total_size, self.mlflow_method]
        row["duration"] = self.duration
        row["duration per file"] = self.duration_per_file
        row["duration per MB"] = self.duration_per_mb
        row["succeeded"] = not self.failed


def run_stress_test(sleep_time: int, dry_run: bool = False):
    results = []
    home = os.getenv("HOME")
    base_dir = pathlib.Path(f"{home}/data/mlflow_server_stress_test")
    values = [1, 10, 100, 1000]
    paths = [
        base_dir / "1MB",
        base_dir / "10MB",
        base_dir / "100MB",
        base_dir / "1000MB",
    ]

    methods = ["log_artifact", "log_artifacts"]

    index = pd.MultiIndex.from_arrays(
        list(zip(*itertools.product(values, methods))),
        names=["total size [MB]", "method"],
    )

    df = pd.DataFrame(
        data=None,
        index=index,
        columns=[
            "duration",
            "duration per MB",
            "duration per file",
            "succeeded",
        ],
    )

    context = stress.get_context(dry_run)

    for total_size, path in zip(values, paths):
        for method in [
            _log_individual_artifacts_for_run,
            _log_grouped_artifacts_for_run,
        ]:
            logger.info(
                "Running test for logging %s MB via %s",
                total_size,
                method.__name__,
            )
            result = method(
                total_size=total_size,
                path=path,
                dry_run=dry_run,
                context=context,
            )
            result.insert_to_df(df)
            results.append(result)
            logger.info(
                "Sleeping for %s seconds until next stress test", sleep_time
            )
            time.sleep(sleep_time)
    df.to_csv("results-artifacts.csv")
    logger.info("Results saved to results-artifacts.csv")
    stress.log_results(
        results,
        columns=[
            "total size [MB]",
            "# files",
            "method",
            "duration",
            "duration per MB",
            "duration per file",
            "succeeded",
        ],
        logger=logger,
    )


def _log_results(results: t.List[Results]) -> None:
    rows = [result.to_table_row() for result in results]
    table = "|  |\n" "| --- | --- | --- | --- | --- | --- |\n" + "".join(rows)
    logger.info("Stress test results: \n%s", table)


def _log_individual_artifacts_for_run(
    total_size: int, path: pathlib.Path, dry_run: bool, context: stress.Context
) -> Results:
    failed = False
    files = list(path.glob("*"))
    n_files = len(files)
    start = time.time()
    with context(run_name=f"test-requests-individual-{total_size}-MB"):
        for i, file in enumerate(files):
            try:
                if not dry_run:
                    mlflow.log_artifact(file.as_posix(), f"test-file-{i}.mp4")
            except:  # noqa: E722
                logger.exception(
                    "Exception after %s individual artifacts", i, exc_info=True
                )
                failed = True
                break
    duration = time.time() - start
    return Results(
        total_size=total_size,
        n_files=n_files,
        mlflow_method="log_artifact",
        duration=duration,
        failed=failed,
    )


def _log_grouped_artifacts_for_run(
    total_size: int,
    path: pathlib.Path,
    dry_run: bool,
    context: stress.Context,
) -> Results:
    failed = False
    files = list(path.glob("*"))
    n_files = len(files)
    start = time.time()
    with context(run_name=f"test-requests-grouped-{total_size}-MB"):
        try:
            if not dry_run:
                mlflow.log_artifacts(path.as_posix())
        except:  # noqa: E722
            logger.exception(
                "Exception sending %s MB grouped artifacts",
                total_size,
                exc_info=True,
            )
            failed = True
    duration = time.time() - start
    return Results(
        total_size=total_size,
        n_files=n_files,
        mlflow_method="log_artifacts",
        duration=duration,
        failed=failed,
    )


if __name__ == "__main__":
    mantik.init_tracking()
    args = stress.setup_argpaser().parse_args()
    if args.verbose:
        logger.setLevel(logging.DEBUG)
    else:
        logger.setLevel(logging.INFO)
    run_stress_test(sleep_time=int(args.sleep_time), dry_run=args.dry_run)



import asyncio
import concurrent.futures as futures
import dataclasses
import itertools
import logging
import os
import pathlib
import time

import mlflow.entities
import pandas as pd

import mantik
import mantik.testing.stress as stress

logger = logging.getLogger()
stress.setup_logger(logger)

N_PARAMS = 50


@dataclasses.dataclass
class Results(stress.ResultsBase):
    n_parallel_runs: int
    duration_per_param: float = None

    def __post_init__(self):
        self.duration_per_param = self.duration / self.n_parallel_runs
        logger.info(
            (
                "Sending %s params via mlflow.%s to %s parallel runs "
                "%s %s seconds, "
                "average duration per param was %s seconds"
            ),
            N_PARAMS,
            self.mlflow_method,
            self.n_parallel_runs,
            "failed after" if self.failed else "succeeded in",
            self.duration,
            self.duration_per_param,
        )

    def to_table_row(self) -> str:
        return (
            f"| {self.n_parallel_runs} "
            f"| {self.mlflow_method} "
            f"| {self.duration} "
            f"| {self.duration_per_param} "
            f"| {not self.failed} "
            f"|\n"
        )

    def insert_to_df(self, df: pd.DataFrame) -> None:
        row = df.loc[self.n_parallel_runs, self.mlflow_method]
        row["duration"] = self.duration
        row["duration per param"] = self.duration_per_param
        row["succeeded"] = not self.failed


async def run_stress_test(loop, sleep_time: int, dry_run: bool = False):
    values = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 100, 1000, 10000]
    methods = ["log_params"]
    index = pd.MultiIndex.from_arrays(
        list(zip(*itertools.product(values, methods))),
        names=["# parallel runs", "method"],
    )
    home = os.getenv("HOME")
    artifacts = pathlib.Path(f"{home}/data/mlflow_server_stress_test/100MB")

    df = pd.DataFrame(
        data=None,
        index=index,
        columns=["duration", "duration per param", "succeeded"],
    )

    context = stress.get_context(dry_run)

    results = []

    for n_parallel_runs in values:
        for method in methods:
            logger.info(
                "Running test for logging to %s parallel runs via %s",
                n_parallel_runs,
                method,
            )
            start = time.time()

            executor = futures.ProcessPoolExecutor(max_workers=n_parallel_runs)

            await asyncio.gather(
                *(
                    loop.run_in_executor(
                        executor,
                        _log_params_for_run,
                        n_parallel_runs,
                        N_PARAMS,
                        artifacts,
                        dry_run,
                        context,
                    )
                    for _ in range(n_parallel_runs)
                )
            )

            duration = time.time() - start
            result = Results(
                n_parallel_runs=n_parallel_runs,
                mlflow_method="log_params",
                duration=duration,
                failed=False,
            )
            result.insert_to_df(df)
            results.append(result)
            logger.info(
                "Sleeping for %s seconds until next stress test", sleep_time
            )
            time.sleep(sleep_time)
    df.to_csv("results-parallel-runs.csv")
    logger.info("Results saved to results-parallel-runs.csv")
    stress.log_results(
        results,
        columns=[
            "# parallel runs",
            "method",
            "duration",
            "duration per param",
            "succeeded",
        ],
        logger=logger,
    )


def _log_params_for_run(
    n_parallel_runs: int,
    n_params: int,
    artifacts: pathlib.Path,
    dry_run: bool,
    context: stress.Context,
) -> None:
    logger.info("Logging %s params and artifacts from %s", n_params, artifacts)
    params = {f"param-{i}": 1 for i in range(n_params)}
    with context(run_name=f"test-requests-grouped-{n_params}"):
        try:
            if not dry_run:
                mlflow.log_params(params)
                mlflow.log_artifacts(artifacts.as_posix())
        except:  # noqa: E722
            logger.exception(
                "Exception sending %s grouped params for %s parallel runs",
                n_params,
                n_parallel_runs,
                exc_info=True,
            )


if __name__ == "__main__":
    mantik.init_tracking()
    args = stress.setup_argpaser().parse_args()
    if args.verbose:
        logger.setLevel(logging.DEBUG)
    else:
        logger.setLevel(logging.INFO)
    loop = asyncio.get_event_loop()
    loop.run_until_complete(
        run_stress_test(
            loop, sleep_time=int(args.sleep_time), dry_run=args.dry_run
        )
    )



"""MLflow local run function.

This source file contains code from [MLflow](https://github.com/mlflow/mlflow)
licensed under Apache-2.0 license, see
[here](https://github.com/mlflow/mlflow/blob/1eef4641df6f605b7d9faa83b0fc25e65877dbf4/LICENSE.txt)
for the original license.

Changes made to the original source code are denoted as such with comments.

For Mantik developers
NOTE: this has to be kept up to date with the different Mlflow versions.

If any changes occur in the original code that can be found here:
https://github.com/mlflow/mlflow/blob/master/mlflow/projects/backend/local.py
"""
import io
import os
import subprocess
import sys

import mlflow


def run_entry_point(command, work_dir, experiment_id, run_id):
    """
    Run an entry point command in a subprocess,
    returning a SubmittedRun that can be used to
    query the run's status.

    Args:
        command: Entry point command to run
        work_dir: Working directory in which to run the command
        run_id: MLflow run ID associated with the entry point execution.
    """
    env = os.environ.copy()
    env.update(mlflow.projects.utils.get_run_env_vars(run_id, experiment_id))
    try:
        env.update(
            mlflow.utils.databricks_utils.get_databricks_env_vars(
                tracking_uri=mlflow.get_tracking_uri()
            )
        )
    except AttributeError:
        env.update(
            mlflow.projects.utils.get_databricks_env_vars(
                tracking_uri=mlflow.get_tracking_uri()
            )
        )
    mlflow.projects.backend.local._logger.info(
        "=== Running command '%s' in run with ID '%s' === ", command, run_id
    )
    # in case os name is not 'nt', we are not running on windows. It introduces
    # bash command otherwise.

    # MANTIK COMMENT
    # Added next line to avoid buffering of output
    # so that we can print it real time
    env["PYTHONUNBUFFERED"] = "1"

    # MANTIK COMMENT
    # Changed next lines to capture output of
    # mantik run subprocess to print it in main process
    # What Changed (All changes refer both to the windows and linux subprocess):
    # - subprocess.Popen called with a context manager
    # - stdout=subprocess.PIPE, stderr=subprocess.STDOUT
    #   and bufsize=-1 added to subprocess.Popen
    # - Added indented lines in the context manager
    #   to print in real time here
    #   in the main process the logs of the subprocess
    if os.name != "nt":
        with subprocess.Popen(
            ["bash", "-c", command],
            close_fds=True,
            cwd=work_dir,
            env=env,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            bufsize=-1,
        ) as process:
            for line in io.TextIOWrapper(process.stdout, newline=""):
                print(line, flush=True, end="")
    else:
        with subprocess.Popen(
            ["cmd", "/c", command],
            close_fds=True,
            cwd=work_dir,
            env=env,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            bufsize=-1,
        ) as process:
            for line in io.TextIOWrapper(process.stderr, newline=""):
                print(line, flush=True, end="")
    return mlflow.projects.submitted_run.LocalSubmittedRun(run_id, process)


def run_mlflow_run_cmd(mlflow_run_arr, env_map):
    """
    Invoke ``mlflow run`` in a subprocess,
    which in turn runs the entry point in a child process.
    Returns a handle to the subprocess. Popen launched to invoke ``mlflow run``.
    """
    final_env = os.environ.copy()
    final_env.update(env_map)
    # Launch `mlflow run` command
    # as the leader of its own process group so that we can do a
    # best-effort cleanup of all its descendant processes if needed

    # MANTIK COMMENT
    # Added next line to avoid buffering of output
    # so that we can print it real time
    final_env["PYTHONUNBUFFERED"] = "1"

    # MANTIK COMMENT
    # Changed next lines to capture output of
    # mantik run subprocess to print it in main process
    # What Changed (All changes refer both to the windows and linux subprocess):
    # - subprocess.Popen called with a context manager
    # - stdout=subprocess.PIPE, stderr=subprocess.STDOUT
    #   and bufsize=-1 added to subprocess.Popen
    # - Added indented lines in the context manager
    #   to print in real time here
    #   in the main process the logs of the subprocess
    # - Remove text=True from subprocess.Popen so that output is a bytestream
    if sys.platform == "win32":
        with subprocess.Popen(
            mlflow_run_arr,
            env=final_env,
            creationflags=subprocess.CREATE_NEW_PROCESS_GROUP,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            bufsize=-1,
        ) as process:
            for line in io.TextIOWrapper(process.stderr, newline=""):
                print(line, flush=True, end="")
        return process
    else:
        with subprocess.Popen(
            mlflow_run_arr,
            env=final_env,
            preexec_fn=os.setsid,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            bufsize=-1,
        ) as process:
            for line in io.TextIOWrapper(process.stderr, newline=""):
                print(line, flush=True, end="")
        return process



import pathlib
import typing as t

import click

MLPROJECT_PATH = click.argument(
    "mlproject-path",
    type=click.Path(path_type=pathlib.Path),
    required=True,
)

ENTRY_POINT = click.option(
    "--entry-point",
    required=False,
    default="main",
    show_default=True,
    help="Entrypoint of the MLproject file.",
)

BRANCH = click.option(
    "--branch",
    default=None,
    type=str,
    help="Name of the code repository's branch you want to use for this run",
    required=True,
)

COMMIT = click.option(
    "--commit",
    default=None,
    type=str,
    help="""Name of the code repository's full commit hash you want to use for
        this run.

        If both branch and commit hash are given, the commit hash is preferred
        over the branch.
    """,
    required=False,
)

PARAMETER = click.option(
    "--parameter", "-P", show_default=True, default=lambda: [], multiple=True
)


def check_commit_or_branch(
    branch: t.Optional[str], commit: t.Optional[str], logger
):
    if branch is None and commit is None:
        raise ValueError(
            "Either provide a branch or full commit hash to submit"
        )
    elif branch is not None and commit is not None:
        logger.warning(
            "Both branch name %s and commit hash %s given, using commit hash to"
            " submit run",
            branch,
            commit,
        )



import logging
import typing as t
import uuid

import fs.base

import mantik.authentication.auth
import mantik.remote_file_service.abstract_file_service as afs
from mantik.remote_file_service.abstract_file_service import FilePath

_PROJECT_ID_ENV_VAR = "MANTIK_PROJECT_ID"

logger = logging.getLogger(__name__)


class DataClientException(Exception):
    pass


class DataClient(afs.AbstractFileService):
    def localise_path(self, target: FilePath) -> FilePath:
        return self.file_service.localise_path(target)

    def is_remote(self, target: FilePath) -> bool:
        return self.file_service.is_remote(target)

    def _get_remote_fs(self, **nargs) -> fs.base.FS:
        return self.file_service._get_remote_fs(**nargs)

    def __init__(
        self,
        file_service: afs.AbstractFileService,
    ):
        self.access_token = mantik.authentication.auth.get_valid_access_token()
        self.file_service = file_service
        super().__init__()

    @classmethod
    def from_env(
        cls,
        connection_id: t.Optional[uuid.UUID] = None,
        remote_fs_type: t.Optional[t.Type[afs.AbstractFileService]] = None,
    ) -> "DataClient":
        if remote_fs_type is None:
            supported_types = _get_available_remote_file_systems()
            remote_fs_type_from_env = mantik.utils.env.get_required_env_var(
                afs.REMOTE_FS_TYPE_ENV_VAR
            )
            try:
                remote_fs_type = supported_types[remote_fs_type_from_env]
            except KeyError:
                raise DataClientException(
                    "Invalid remote file system type, set REMOTE_FS_TYPE"
                    f"as one of this supported types {list(supported_types.keys())}"  # noqa
                )

        return cls(
            file_service=remote_fs_type.from_env(connection_id=connection_id),
        )

    def list_directory(
        self, target: afs.FilePath
    ) -> t.List[t.Union[afs.Directory, afs.File]]:
        return self.file_service.list_directory(target)

    def create_directory(
        self,
        target: afs.FilePath,
    ) -> afs.Directory:
        return self.file_service.create_directory(target)

    def remove_directory(
        self,
        target: afs.FilePath,
    ) -> None:
        self.file_service.remove_directory(target)

    def copy_directory(
        self,
        source: afs.FilePath,
        target: afs.FilePath,
    ) -> afs.Directory:
        return self.file_service.copy_directory(source, target)

    def create_file_if_not_exists(
        self,
        target=afs.FilePath,
    ) -> afs.File:
        return self.file_service.create_file_if_not_exists(target)

    def remove_file(
        self,
        target: afs.FilePath,
    ) -> None:
        self.file_service.remove_file(target)

    def copy_file(
        self,
        source: afs.FilePath,
        target: afs.FilePath,
    ) -> afs.File:
        return self.file_service.copy_file(source, target)

    def exists(self, target=afs.FilePath) -> bool:
        return self.file_service.exists(target=target)

    @property
    def user(self) -> str:
        return self.file_service.user

    def change_permissions(
        self, target: afs.FilePath, new_permissions: afs.FileMeta
    ) -> None:
        self.file_service.change_permissions(target, new_permissions)


def _str_or_uuid_to_uuid(id_: t.Union[str, uuid.UUID]):
    if isinstance(id_, str):
        try:
            id_ = uuid.UUID(id_)
        except ValueError:
            raise DataClientException(
                "Badly formed hexadecimal UUID string for project ID"
            )
    return id_


def _get_available_remote_file_systems() -> t.Dict:
    supported_types = {}
    _attempt_file_service_import(
        supported_types=supported_types,
        name="S3",
        module="s3_file_service",
        cls="S3FileService",
        extras="s3",
    )
    _attempt_file_service_import(
        supported_types=supported_types,
        name="UNICORE",
        module="unicore_file_service",
        cls="UnicoreFileService",
        extras="s3",
    )
    return supported_types


def _attempt_file_service_import(
    supported_types: t.Dict, name: str, module: str, cls: str, extras: str
) -> None:
    """Attempt to import file service from module.

    Parameters
    ----------
    supported_types : dict
        The supported remote file system types.

        If available, the file system will be added.
    name : str
        Name of the remote file system type.

        E.g. ``"S3"`` or ``"UNICORE"``.
    module : str
        Name of the module inside ``mantik.remote_file_system``.

        E.g. ``"s3_file_service"`` or ``"unicore_file_service"``.
    cls : str
        Name of the class inside the respective module.
    extras : str
        Name of the extras inside ``pyproject.toml`` required for
        the remote file system.

    Notes
    -----
    This is required to avoid import errors when extras are not installed.

    Extras for the respective file services are defined in ``pyproject.toml``.

    """
    try:
        sub_module = getattr(mantik.remote_file_service, module)
    except ModuleNotFoundError:
        logger.warning(
            (
                "Unable to import %s remote file service, "
                'install required requirements via `pip install "mantik[%s]"`'
            ),
            name,
            extras,
            exc_info=True,
        )
    else:
        supported_types[name] = getattr(sub_module, cls)



import uuid

import click

import mantik.cli._options as _options
import mantik.cli.main as main
import mantik.runs.artifacts as artifacts


@main.cli.group("runs")
def cli() -> None:
    """Interaction with Mantik runs."""


@cli.command("download-artifacts")
@_options.PROJECT_ID
@_options.get_run_id(required=True)
@_options.get_target_dir_option(
    help_option="Path to directory where the artifacts will be downloaded.",
)
@click.option(
    "--unzip/--no-unzip",
    default=True,
    help="Unzip the artifacts in the target dir.",
    show_default=True,
)
def download_artifacts(
    project_id: uuid.UUID,
    run_id: uuid.UUID,
    target_dir: str,
    unzip: bool,
) -> None:
    """Download the artifacts from a run."""
    artifacts.download_artifacts(
        project_id=project_id, run_id=run_id, target_dir=target_dir, unzip=unzip
    )



import os
import uuid

import pytest

import mantik.runs.artifacts as artifacts


@pytest.mark.parametrize("unzip", [True, False])
def test_download_artifacts(
    mock_authentication,
    tmpdir,
    mock_get_artifacts_url,
    mock_get_url,
    info_caplog,
    unzip,
    zipped_file_name,
) -> None:
    project_id = uuid.uuid4()
    run_id = uuid.uuid4()
    with mock_get_artifacts_url as get_artifacts_url, mock_get_url as get_url:
        artifacts.download_artifacts(
            project_id=project_id,
            run_id=run_id,
            target_dir=str(tmpdir),
            unzip=unzip,
        )

        get_artifacts_url.assert_called()
        get_url.assert_called()
        if not unzip:
            assert (
                f"Successfully downloaded at {str(tmpdir)}/artifacts.zip"
                in info_caplog.text
            )
        else:
            assert (
                f"Artifacts successfully downloaded at {str(tmpdir)}"
                in info_caplog.text
            )
            assert zipped_file_name in os.listdir(tmpdir)



import mantik_compute_backend.unicore.exceptions as exceptions


def test_unicore_exception():
    exc = exceptions.UnicoreError("Test")
    assert str(exc) == "Test"


def test_authentication_failed_exception():
    exc = exceptions.AuthenticationFailedException("Test")
    assert str(exc) == "Test"
    # Make sure the error inherits from UnicoreError
    assert exceptions.UnicoreError in exc.__class__.__bases__



import glob
import pathlib
import tempfile
import unittest.mock
import uuid

import mantik.testing as testing
import mantik.utils as utils
import mantik.utils.credentials as _credentials
import mantik_compute_backend.handle_submit_run as backend


@unittest.mock.patch("mlflow.projects.run")
def test_handle_submit_run_request(
    mock_run, submit_run_request_data, submit_run_request_files
):
    # MLflow returns run IDs as UUID string without dashes (`uuid.hex`)
    mock_run.return_value.run_id = uuid.uuid4().hex
    mock_run.return_value.job_id = "2"
    with utils.env.env_vars_set({utils.mlflow.TRACKING_URI_ENV_VAR: "foo"}):
        response = backend.handle_submit_run_request(
            **submit_run_request_data, **submit_run_request_files
        )
        assert response
        mock_run.assert_called()
        # Assert that secret environment variables are unset
        testing.env.assert_env_var(_credentials.UNICORE_USERNAME_ENV_VAR, None)


def test_unzip_to_tmp(zipped_unicore_content):
    with tempfile.TemporaryDirectory() as directory:
        backend._unzip_to_file(zipped_unicore_content, directory)
        files = list(
            map(
                lambda filepath: pathlib.Path(filepath).name,
                glob.glob(directory + "/*.py"),
            )
        )
        assert "main.py" in files



import click

# TODO this two arguments could be replaced by using the
#      get_target_dir_option from mantik/cli/_options.py
#      and something similar for source

SOURCE = click.argument(
    "source",
    type=str,
    nargs=1,
    required=True,
)
TARGET = click.argument(
    "target",
    type=str,
    nargs=1,
    required=True,
)



import logging
import sys

import mlflow

logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.DEBUG)

if __name__ == "__main__":
    # Print, log and write to file to test all those functionalities with the
    # UNICORE backend.
    argument_string = " ".join(sys.argv[1:])
    with mlflow.start_run():
        print(argument_string)
        logger.debug(f"Printed {argument_string}.")
        mlflow.log_param("print", "Hello")
        with open("output.txt", "w") as text_file:
            text_file.write("test")



import pathlib
import shutil

import boto3
import fs_s3fs
import moto
import pytest

import mantik.remote_file_service.s3_file_service as s3_fs
import mantik.utils.env as env

BUCKET_NAME = "mocked-bucket"

LOCAL_WORKING_DIRECTORY = pathlib.Path(__file__).parent / "workspace"

SAMPLE_FOLDER = "sample-directory"
SAMPLE_FILENAME = "hello-world.txt"
SAMPLE_FILE_FILE_PATH = f"{SAMPLE_FOLDER}/{SAMPLE_FILENAME}"

DOWNLOADS = "downloads"


@pytest.fixture
def mock_s3():
    with moto.mock_s3():
        client = boto3.client("s3")
        client.create_bucket(
            Bucket=BUCKET_NAME,
            CreateBucketConfiguration={"LocationConstraint": "eu-central-1"},
        )

        yield


@pytest.fixture
def mock_s3_with_sample_file(mock_s3):
    client = boto3.client("s3")
    client.upload_file(
        f"{LOCAL_WORKING_DIRECTORY}/{SAMPLE_FILE_FILE_PATH}",
        BUCKET_NAME,
        SAMPLE_FILE_FILE_PATH,
    )
    fs_s3fs.S3FS(
        bucket_name=BUCKET_NAME,
        aws_access_key_id="1234",
        aws_secret_access_key="XXXX",
    ).makedir(
        SAMPLE_FOLDER
    )  # Needed because there is no 'real' concept of folders in S3
    # See: https://github.com/PyFilesystem/s3fs/issues/17
    yield


@pytest.fixture
def temp_testing_folder_path(tmp_path) -> str:
    # Copy the source folder to the temporary directory
    shutil.copytree(
        f"{LOCAL_WORKING_DIRECTORY}/{SAMPLE_FOLDER}", tmp_path / SAMPLE_FOLDER
    )
    # Create a folder to download to, in the new temp directory
    pathlib.Path(tmp_path / DOWNLOADS).mkdir()

    yield str(tmp_path)


@pytest.fixture()
def s3_fs_client(temp_testing_folder_path) -> s3_fs.S3FileService:
    return s3_fs.S3FileService(
        aws_access_key_id="1234",
        aws_secret_access_key="XXXX",
        local_base_path=temp_testing_folder_path,
    )


def test_list_directory(s3_fs_client: s3_fs, mock_s3_with_sample_file) -> None:
    local_list_dir = s3_fs_client.list_directory(SAMPLE_FOLDER)
    assert len(local_list_dir) == 1
    assert local_list_dir[0].path == SAMPLE_FILE_FILE_PATH
    remote_list_dir = s3_fs_client.list_directory(
        f"s3://{BUCKET_NAME}/{SAMPLE_FOLDER}"
    )
    assert len(remote_list_dir) == 1
    assert (
        remote_list_dir[0].path == f"s3://{BUCKET_NAME}/{SAMPLE_FILE_FILE_PATH}"
    )


def test_copy_file_both_ways_works(
    s3_fs_client: s3_fs,
    mock_s3,
) -> None:
    # from local to s3
    assert not s3_fs_client.exists(f"s3://{BUCKET_NAME}/{SAMPLE_FILENAME}")

    s3_fs_client.copy_file(
        f"{SAMPLE_FILE_FILE_PATH}",
        f"s3://{BUCKET_NAME}/{SAMPLE_FILENAME}",
    )
    assert s3_fs_client.exists(f"s3://{BUCKET_NAME}/{SAMPLE_FILENAME}")

    # from s3 to local
    assert not s3_fs_client.exists(f"{DOWNLOADS}/{SAMPLE_FILENAME}")

    s3_fs_client.copy_file(
        f"s3://{BUCKET_NAME}/{SAMPLE_FILENAME}",
        f"{DOWNLOADS}/{SAMPLE_FILENAME}",
    )
    assert s3_fs_client.exists(f"{DOWNLOADS}/{SAMPLE_FILENAME}")


def test_remove_file(
    s3_fs_client: s3_fs,
    mock_s3_with_sample_file,
) -> None:
    # s3
    assert s3_fs_client.exists(f"s3://{BUCKET_NAME}/{SAMPLE_FILE_FILE_PATH}")
    s3_fs_client.remove_file(f"s3://{BUCKET_NAME}/{SAMPLE_FILE_FILE_PATH}")
    assert not s3_fs_client.exists(
        f"s3://{BUCKET_NAME}/{SAMPLE_FILE_FILE_PATH}"
    )

    #  local
    assert s3_fs_client.exists(SAMPLE_FILE_FILE_PATH)
    s3_fs_client.remove_file(SAMPLE_FILE_FILE_PATH)
    assert not s3_fs_client.exists(SAMPLE_FILE_FILE_PATH)


def test_copy_directory_both_ways_works(s3_fs_client: s3_fs, mock_s3) -> None:
    # from local to s3
    assert not s3_fs_client.exists(f"s3://{BUCKET_NAME}/{SAMPLE_FOLDER}")

    s3_fs_client.copy_directory(
        SAMPLE_FOLDER, f"s3://{BUCKET_NAME}/{SAMPLE_FOLDER}"
    )
    assert s3_fs_client.exists(f"s3://{BUCKET_NAME}/{SAMPLE_FOLDER}")

    assert len(
        s3_fs_client.list_directory(f"s3://{BUCKET_NAME}/{SAMPLE_FOLDER}")
    ) == len(s3_fs_client.list_directory(SAMPLE_FOLDER))

    # from s3 to local
    assert not s3_fs_client.exists(f"{DOWNLOADS}/{SAMPLE_FOLDER}")

    s3_fs_client.copy_directory(
        f"s3://{BUCKET_NAME}/{SAMPLE_FOLDER}", f"{DOWNLOADS}/{SAMPLE_FOLDER}"
    )
    assert s3_fs_client.exists(f"{DOWNLOADS}/{SAMPLE_FOLDER}")

    assert len(
        s3_fs_client.list_directory(f"{DOWNLOADS}/{SAMPLE_FOLDER}")
    ) == len(s3_fs_client.list_directory(SAMPLE_FOLDER))


def test_remove_directory(
    s3_fs_client: s3_fs,
    mock_s3_with_sample_file,
) -> None:
    # s3
    assert s3_fs_client.exists(f"s3://{BUCKET_NAME}/{SAMPLE_FOLDER}")
    s3_fs_client.remove_directory(f"s3://{BUCKET_NAME}/{SAMPLE_FOLDER}")
    assert not s3_fs_client.exists(f"s3://{BUCKET_NAME}/{SAMPLE_FOLDER}")

    #  local
    assert s3_fs_client.exists(SAMPLE_FOLDER)
    s3_fs_client.remove_directory(SAMPLE_FOLDER)
    assert not s3_fs_client.exists(SAMPLE_FOLDER)


def test_create_file_if_not_exists(s3_fs_client: s3_fs, mock_s3) -> None:
    s3_fs_client.create_file_if_not_exists(f"s3://{BUCKET_NAME}/new-file.txt")
    assert s3_fs_client.exists(f"s3://{BUCKET_NAME}/new-file.txt")


def test_create_directory(s3_fs_client: s3_fs, mock_s3) -> None:
    s3_fs_client.create_directory(f"s3://{BUCKET_NAME}/new-folder")
    assert s3_fs_client.exists(f"s3://{BUCKET_NAME}/new-folder")


def test_user(s3_fs_client: s3_fs) -> None:
    assert s3_fs_client.user == "1234"


def test_init_from_env_vars() -> None:
    with env.env_vars_set(
        {
            s3_fs._USERNAME_ENV_VAR: "env-username",
            s3_fs._PASSWORD_ENV_VAR: "env-password",
        }
    ):
        client = s3_fs.S3FileService.from_env()
        assert client.aws_access_key_id == "env-username"
        assert client.aws_secret_access_key == "env-password"


def test_init_from_connection_id(
    mock_api_creds_retrieval,
    fake_connection,
    fake_connection_id,
) -> None:
    with env.env_vars_set({"MANTIK_USERNAME": "X", "MANTIK_PASSWORD": "X"}):
        client = s3_fs.S3FileService.from_env(connection_id=fake_connection_id)
        assert client.aws_access_key_id == fake_connection.login_name
        assert client.aws_secret_access_key == fake_connection.password



import freezegun
import pytest

import mantik.testing as testing
import tokens.cognito as _cognito
import tokens.exceptions as exceptions
import tokens.verifier as _verifier


@pytest.fixture()
def verifier() -> _verifier.TokenVerifier:
    cognito = _cognito.client.Properties(
        region=testing.mlflow_server.FAKE_COGNITO_REGION,
        user_pool_id=testing.mlflow_server.FAKE_COGNITO_USER_POOL_ID,
        app_client_id=testing.mlflow_server.FAKE_COGNITO_APP_CLIENT_ID,
        app_client_secret=testing.mlflow_server.FAKE_COGNITO_APP_CLIENT_SECRET,
    )
    jwks = testing.mlflow_server.FakeJWKS.create()
    return _verifier.TokenVerifier(
        cognito=cognito,
        jwks=jwks,
    )


class TestTokenVerifier:
    @pytest.mark.parametrize(
        ("current_time", "token", "expected"),
        [
            (
                testing.mlflow_server.NON_EXPIRED_DATETIME,
                testing.mlflow_server.FAKE_JWT,
                None,
            ),
            (
                testing.mlflow_server.EXPIRED_DATETIME,
                testing.mlflow_server.FAKE_JWT,
                exceptions.TokenExpiredException(),
            ),
            (
                testing.mlflow_server.NON_EXPIRED_DATETIME,
                testing.mlflow_server.FAKE_JWT_INVALID_SIGNATURE,
                exceptions.InvalidSignatureException(),
            ),
            (
                testing.mlflow_server.NON_EXPIRED_DATETIME,
                testing.mlflow_server.FAKE_JWT_INVALID_CLIENT_ID,
                # Currently unable to test for the expected dedicated exception
                # (`exceptions.InvalidClientException`) since it seems to be
                # impossible to change the payload without breaking the
                # signature.
                exceptions.InvalidSignatureException(),
            ),
            (
                testing.mlflow_server.NON_EXPIRED_DATETIME,
                testing.mlflow_server.FAKE_JWT_INVALID_ISSUER,
                # See above, dedicated exception here is
                # `exceptions.InvalidIssuerException`
                exceptions.InvalidSignatureException(),
            ),
            (
                testing.mlflow_server.NON_EXPIRED_DATETIME,
                testing.mlflow_server.FAKE_JWT_INVALID_TOKEN_TYPE,
                # See above, dedicated exception here is
                # `exceptions.IncorrectTokenTypeException`
                exceptions.InvalidSignatureException(),
            ),
        ],
    )
    def test_verify_token(
        self, expect_raise_if_exception, verifier, current_time, token, expected
    ):
        with freezegun.freeze_time(current_time):
            with expect_raise_if_exception(expected):
                verifier.verify_token(token)
